<?php

namespace QUI\UserDownloads;

use QUI;
use QUI\Interfaces\Users\User as UserInterface;
use QUI\Projects\Media\Item as QUIMediaItem;
use QUI\Projects\Media\File as QUIMediaFile;
use QUI\Projects\Media\Image as QUIMediaImage;
use QUI\Projects\Media\Folder as QUIMediaFolder;
use QUI\Projects\Media\Utils as QUIMediaUtils;

use function array_key_exists;
use function array_search;
use function array_unique;
use function array_values;
use function basename;
use function current;
use function date_create;
use function hash;
use function implode;
use function in_array;
use function json_encode;
use function mb_strpos;
use function usort;

/**
 * Class DownloadEntry
 *
 * Represents an entry for downloadable content.
 *
 * Download entries are displayed in the frontend profile
 * and can be accessed by the assigned user.
 *
 * Examples:
 *   - An order that contains download products with corresponding files
 *   - A collection of QUIQQER media items (folders / items)
 *   - A collection of custom files (e.g. documents, tickets)
 *   - A collection of download URLs
 */
class DownloadEntry
{
    const CUSTOM_DATA_KEY_DOWNLOAD_MEMBERSHIP_PRODUCT_IDS = 'downloadMembershipProductIds';

    protected int | string | null $id;

    protected QUI\Interfaces\Users\User $User;

    protected array $title = [];

    protected array $description = [];

    private array $customData = [];

    /**
     * Media items accessible via this download entry
     *
     * @var string[]
     */
    protected array $quiqqerMediaUrls = [];

    /**
     * URLs with (optional) titles that are accessible
     *
     * @var array
     */
    protected array $urls = [];

    /**
     * @var string|null
     */
    protected ?string $orderHash = null;

    /**
     * DownloadEntry constructor.
     *
     * @param UserInterface $User
     * @param int|string|null $id (optional) - Database id
     */
    public function __construct(UserInterface $User, null | int | string $id = null)
    {
        $this->User = $User;
        $this->id = $id;
    }

    /**
     * @return int|string|null
     */
    public function getId(): null | int | string
    {
        return $this->id;
    }

    /**
     * @return UserInterface
     */
    public function getUser(): UserInterface
    {
        return $this->User;
    }

    /**
     * @param QUI\Locale|null $Locale (optional)
     * @return string
     */
    public function getTitle(?QUI\Locale $Locale = null): string
    {
        if (empty($Locale)) {
            $Locale = QUI::getLocale();
        }

        if (!empty($this->title[$Locale->getCurrent()])) {
            return $this->title[$Locale->getCurrent()];
        }

        return '';
    }

    /**
     * @param string $lang
     * @param string $title
     */
    public function setTitle(string $lang, string $title): void
    {
        $this->title[$lang] = $title;
    }

    /**
     * @param QUI\Locale|null $Locale (optional)
     * @return string
     */
    public function getDescription(?QUI\Locale $Locale = null): string
    {
        if (empty($Locale)) {
            $Locale = QUI::getLocale();
        }

        if (!empty($this->description[$Locale->getCurrent()])) {
            return $this->description[$Locale->getCurrent()];
        }

        return '';
    }

    /**
     * @param string $lang
     * @param string $description
     */
    public function setDescription(string $lang, string $description): void
    {
        $this->description[$lang] = $description;
    }

    /**
     * @return string|null
     */
    public function getOrderHash(): ?string
    {
        return $this->orderHash;
    }

    /**
     * @param string|null $orderHash
     */
    public function setOrderHash(?string $orderHash): void
    {
        $this->orderHash = $orderHash;
    }

    /**
     * If an order is associated with this download entry check if the order is still valid,
     * meaning it has a status that is not considered "cancelled" or otherwise reverted.
     *
     * @return bool
     */
    public function isOrderValid(): bool
    {
        if (empty($this->orderHash)) {
            return true;
        }

        if (!class_exists('QUI\ERP\Order\Handler')) {
            return false;
        }

        try {
            $Order = QUI\ERP\Order\Handler::getInstance()->getOrderByHash($this->orderHash);
        } catch (\Exception $Exception) {
            if ($Exception->getCode() === QUI\ERP\Order\Handler::ERROR_ORDER_NOT_FOUND) {
                return false;
            }

            QUI\System\Log::writeException($Exception);
            return true;
        }

        $settings = Handler::getSettings();

        if (empty($settings['orderCancelStatus'])) {
            return true;
        }

        return $Order->getProcessingStatus()->getId() !== (int)$settings['orderCancelStatus'];
    }

    // region QUIQQER Media Items

    /**
     * Adds a QUIQQER media item to the download entry.
     *
     * Depending on the type of item, this means the following things:
     *
     *  - Media folder -> All files in the folder will be downloadble (NOT parsed recursively!; single level only; active items only)
     *  - Media file -> This file will be downloadable (active only)
     *  - Media image -> This image file will be downloadable (active only)
     *
     * @param QUIMediaItem $MediaItem
     * @return void
     */
    public function addMediaItem(QUIMediaItem $MediaItem): void
    {
        if ($MediaItem instanceof QUIMediaFolder) {
            /** @var QUIMediaItem $Child */
            foreach ($MediaItem->getChildren() as $Child) {
                if ($Child instanceof QUIMediaFolder) {
                    continue;
                }

                if ($Child->isDeleted()) {
                    continue;
                }

                // Currently inactive items can be downloaded, since active items are always downloadable
                // but should not be in case of buyable downloads
//                if (!$Child->isActive()) {
//                    continue;
//                }

                $this->quiqqerMediaUrls[] = $Child->getUrl();
            }
        } elseif (!$MediaItem->isDeleted()) {
            $this->quiqqerMediaUrls[] = $MediaItem->getUrl();
        } else {
            return;
        }

        $this->quiqqerMediaUrls = array_values(array_unique($this->quiqqerMediaUrls));
    }

    /**
     * Remove a QUIQQER media item from this DownloadEntry
     *
     * @param QUIMediaItem $MediaItem
     * @return void
     */
    public function removeMediaItem(QUIMediaItem $MediaItem): void
    {
        if ($MediaItem instanceof QUIMediaFolder) {
            /** @var QUIMediaItem $Child */
            foreach ($MediaItem->getChildren() as $Child) {
                if ($Child instanceof QUIMediaFolder) {
                    continue;
                }

                unset($this->quiqqerMediaUrls[array_search($Child->getUrl(), $this->quiqqerMediaUrls)]);
            }
        } else {
            unset($this->quiqqerMediaUrls[array_search($MediaItem->getUrl(), $this->quiqqerMediaUrls)]);
        }

        $this->quiqqerMediaUrls = array_values($this->quiqqerMediaUrls);
    }

    /**
     * Add a quiqqer media items URL to this download entry.
     *
     * @param string $mediaUrl
     * @throws Exception
     */
    public function addQuiqqerMediaItemUrl(string $mediaUrl): void
    {
        if (!QUIMediaUtils::isMediaUrl($mediaUrl)) {
            throw new Exception(
                'Url "' . $mediaUrl . '" ist no a valid QUIQQER media url!'
            );
        }

        $this->quiqqerMediaUrls[] = $mediaUrl;
        $this->quiqqerMediaUrls = array_values(array_unique($this->quiqqerMediaUrls));
    }

    /**
     * Remove a specific QUIQQER media item URL from this download entry.
     *
     * @param string $mediaUrl
     * @throws Exception
     */
    public function removeQuiqqerMediaItemUrl(string $mediaUrl): void
    {
        if (!QUIMediaUtils::isMediaUrl($mediaUrl)) {
            throw new Exception(
                'Url "' . $mediaUrl . '" ist no a valid QUIQQER media url!'
            );
        }

        if (!$mediaUrl) {
            return;
        }

        if (!in_array($mediaUrl, $this->quiqqerMediaUrls)) {
            return;
        }

        unset($this->quiqqerMediaUrls[array_search($mediaUrl, $this->quiqqerMediaUrls)]);
        $this->quiqqerMediaUrls = array_values($this->quiqqerMediaUrls);
    }

    /**
     * @param string[] $quiqqerMediaUrls
     */
    public function setQuiqqerMediaUrls(array $quiqqerMediaUrls): void
    {
        $this->quiqqerMediaUrls = $quiqqerMediaUrls;
    }

    /**
     * @return string[]
     */
    public function getQuiqqerMediaUrls(): array
    {
        return $this->quiqqerMediaUrls;
    }

    // endregion

    // region Custom URLs

    /**
     * @return array
     */
    public function getUrls(): array
    {
        return $this->urls;
    }

    /**
     * @param array $urls
     */
    public function setUrls(array $urls): void
    {
        $this->urls = $urls;
    }

    /**
     * Add custom download URL to this entry
     *
     * @param string $url
     * @param array|null $title (optional) - Array with title per language
     * @return void
     *
     * @throws Exception
     */
    public function addUrl(string $url, ?array $title = null): void
    {
        $allowedProtocols = [
            'http',
            'https',
            'ftp',
            'sftp'
        ];

        $valid = false;

        if (mb_strpos($url, '/') === 0) {
            $valid = true;
        } else {
            foreach ($allowedProtocols as $protocol) {
                if (mb_strpos($url, $protocol . '://') === 0) {
                    $valid = true;
                    break;
                }
            }
        }

        if (!$valid) {
            throw new Exception([
                'quiqqer/user-downloads',
                'exception.DownloadEntry.addUrl.invalid_protocol',
                [
                    'allowedProtocols' => implode(', ', $allowedProtocols)
                ]
            ]);
        }

        $this->urls[] = [
            'title' => $title,
            'url' => $url
        ];
    }

    /**
     * Remove custom URL from this DownloadEntry
     *
     * @param string $url
     * @return void
     */
    public function removeUrl(string $url): void
    {
        foreach ($this->urls as $k => $entry) {
            if ($entry['url'] === $url) {
                unset($this->urls[$k]);
                break;
            }
        }

        $this->urls = array_values($this->urls);
    }

    // endregion

    /**
     * Get number of QUIQQER media items that are no longer available
     *
     * @return int
     */
    public function getMissingFileCount(): int
    {
        $missing = 0;

        foreach ($this->quiqqerMediaUrls as $quiqqerMediaUrl) {
            try {
                $MediaItem = QUIMediaUtils::getMediaItemByUrl($quiqqerMediaUrl);

                if ($MediaItem->isDeleted()) {
                    $missing++;
                }
            } catch (\Exception $Exception) {
                if ($Exception->getCode() === 404) {
                    $missing++;
                }
            }
        }

        return $missing;
    }

    /**
     * Update this download entry.
     *
     * @return void
     *
     * @throws Exception
     * @throws QUI\Database\Exception
     */
    public function update(): void
    {
        if (empty($this->id)) {
            throw new Exception(
                'Cannot update this DownloadEntry because it has no ID.'
            );
        }

        QUI::getDataBase()->update(
            Handler::getTableDownloadEntries(),
            $this->toDatabaseArray(),
            [
                'id' => $this->id
            ]
        );
    }

    /**
     * Permanently delete this download entry.
     *
     * @return void
     *
     * @throws Exception
     * @throws QUI\Database\Exception
     */
    public function delete(): void
    {
        if (empty($this->id)) {
            throw new Exception(
                'Cannot delete this DownloadEntry because it has no ID.'
            );
        }

        Handler::deleteDownloadEntry($this->id);
    }

    /**
     * Get array representation of this download entry
     *
     * @return array
     */
    public function toArray(): array
    {
        return [
            'id' => $this->id,
            'user_id' => $this->User->getId(),
            'order_hash' => $this->orderHash,
            'title' => $this->title,
            'description' => $this->description,
            'custom_data' => $this->getCustomData(),
            'download_items' => [
                'quiqqerFiles' => $this->quiqqerMediaUrls,
                'urls' => $this->urls
            ]
        ];
    }

    /**
     * Get array for database insert
     *
     * @return array
     */
    public function toDatabaseArray(): array
    {
        return [
            'user_id' => $this->User->getId(),
            'order_hash' => $this->orderHash,
            'title' => json_encode($this->title),
            'description' => json_encode($this->description),
            'custom_data' => json_encode($this->customData),
            'download_items' => json_encode([
                'quiqqerFiles' => $this->quiqqerMediaUrls,
                'urls' => $this->urls
            ])
        ];
    }

    /**
     * Get list of all download URLs (QUIQQER files and custom URLs) for frontend
     *
     * @return array<int, array{title: string, url: string}>
     */
    public function getFrontendDownloadUrls(): array
    {
        $downloadUrls = [];
        $Locale = QUI::getLocale();
        $lang = $Locale->getCurrent();

        $mediaItems = [];

        foreach ($this->quiqqerMediaUrls as $quiqqerMediaUrl) {
            try {
                $MediaItem = QUIMediaUtils::getMediaItemByUrl($quiqqerMediaUrl);
            } catch (\Exception $Exception) {
                if ($Exception->getCode() !== 404) {
                    QUI\System\Log::writeException($Exception);
                }

                continue;
            }

            if ($MediaItem->isDeleted()) {
                continue;
            }

            $mediaItems[] = [
                'item' => $MediaItem,
                'quiqqerMediaUrl' => $quiqqerMediaUrl
            ];
        }

        usort($mediaItems, function ($itemA, $itemB) {
            $MediaItemA = $itemA['item'];
            $MediaItemB = $itemB['item'];

            /**
             * @var QUIMediaItem $MediaItemA
             * @var QUIMediaItem $MediaItemB
             */
            $editDateA = date_create($MediaItemA->getAttribute('e_date'));
            $editDateB = date_create($MediaItemB->getAttribute('e_date'));

            if ($editDateA === $editDateB) {
                return 0;
            }

            return $editDateA > $editDateB ? -1 : 1;
        });

        foreach ($mediaItems as $entry) {
            /** @var QUIMediaItem $MediaItem */
            $MediaItem = $entry['item'];
            $hash = hash('sha256', $entry['quiqqerMediaUrl']);

            $downloadUrls[] = [
                'title' => $MediaItem->getTitle($Locale) ?: basename($MediaItem->getPath()),
                'url' => URL_OPT_DIR . 'quiqqer/user-downloads/bin/download.php?hash=' . $hash
            ];
        }

        foreach ($this->urls as $entry) {
            if (!empty($entry['title'][$lang])) {
                $title = $entry['title'][$lang];
            } else {
                $title = current($entry['title']);
            }

            if (empty($title)) {
                $title = $entry['url'];
            }

            $downloadUrls[] = [
                'title' => $title,
                'url' => $entry['url']
            ];
        }

        return $downloadUrls;
    }

    // region Custom data

    /**
     * @return array
     */
    public function getCustomData(): array
    {
        return $this->customData;
    }

    /**
     * @param array $customData
     */
    public function setCustomData(array $customData): void
    {
        $this->customData = $customData;
    }

    /**
     * @param string $key
     * @return mixed
     */
    public function getCustomDataEntry(string $key): mixed
    {
        if (!array_key_exists($key, $this->customData)) {
            return null;
        }

        return $this->customData[$key];
    }

    /**
     * @param string $key
     * @param mixed $value
     * @return void
     */
    public function setCustomDataEntry(string $key, mixed $value): void
    {
        $this->customData[$key] = $value;
    }

    // endregion
}
