<?php

namespace QUI\UserDownloads;

use QUI;
use QUI\ERP\Products\Handler\Products;
use QUI\ERP\Order\AbstractOrder;
use QUI\ERP\Products\Handler\Products as ProductsHandler;
use QUI\ERP\Products\Product\Exception;
use QUI\Interfaces\Users\User;
use QUI\ERP\Products\Field\Types\Folder as FieldTypeFolder;
use QUI\Memberships\Handler as MembershipsHandler;
use QUI\Memberships\Users\Handler as MembershipUsersHandler;
use QUI\Projects\Media\Folder as QUIMediaFolder;

use function array_merge;
use function date;
use function in_array;
use function is_numeric;
use function json_decode;
use function strtotime;

/**
 * Class Handler
 *
 * Main handler for quiqqer/user-downloads
 */
class Handler
{
    /**
     * Tables
     */
    const TBL_DOWNLOAD_ENTRIES = 'user_download_entries';

    /**
     * Product fields
     */
    const PRODUCT_FIELD_ID_DOWNLOAD_FILES = 350;

    /**
     * Runtime cache for module settings
     *
     * @var array
     */
    protected static array $settings = [];

    /**
     * Create download entries from an order that contains download products.
     *
     * @param AbstractOrder $Order
     * @return DownloadEntry[]
     */
    public static function createDownloadEntriesFromOrder(AbstractOrder $Order): array
    {
        $downloadEntries = [];
        $Customer = $Order->getCustomer();
        $orderHash = $Order->getUUID();
        $orderDate = $Order->getCreateDate();
        $orderId = $Order->getPrefixedId();
        $L = new QUI\Locale();

        // Parse download media folder of order products

        foreach ($Order->getArticles() as $Article) {
            try {
                // Do not parse coupon codes / discounts
                if (empty($Article->getId()) || !is_numeric($Article->getId())) {
                    continue;
                }

                $Product = Products::getProduct($Article->getId());

                // Only parse download products
                if (
                    !($Product instanceof DownloadProductType) &&
                    !($Product instanceof DownloadMembershipProductType)
                ) {
                    continue;
                }

                $DownloadEntry = new DownloadEntry($Customer);
                $DownloadEntry->setOrderHash($orderHash);

                $langs = QUI::availableLanguages();

                foreach ($langs as $lang) {
                    $dateFormat = QUI\ERP\Defaults::getDateFormat($lang);
                    $L->setCurrent($lang);

                    $DownloadEntry->setTitle(
                        $lang,
                        $L->get(
                            'quiqqer/user-downloads',
                            'downloadEntry.product.title',
                            [
                                'productTitle' => $Product->getTitle($L),
                            ]
                        )
                    );

                    $DownloadEntry->setDescription(
                        $lang,
                        $L->get(
                            'quiqqer/user-downloads',
                            'downloadEntry.product.description',
                            [
                                'orderDate' => $L->formatDate(strtotime($orderDate), $dateFormat),
                                'orderId' => $orderId
                            ]
                        )
                    );
                }

                /** @var FieldTypeFolder $FolderField */
                $FolderField = $Product->getField(self::PRODUCT_FIELD_ID_DOWNLOAD_FILES);
                $MediaFolder = $FolderField->getMediaFolder();

                if ($MediaFolder) {
                    $DownloadEntry->addMediaItem($MediaFolder);
                }

                $downloadEntryId = self::addDownloadEntry($DownloadEntry);
                $DownloadEntry = self::getDownloadEntryById($downloadEntryId);

                if ($Product instanceof DownloadMembershipProductType) {
                    $productIds = $DownloadEntry->getCustomDataEntry(
                        DownloadEntry::CUSTOM_DATA_KEY_DOWNLOAD_MEMBERSHIP_PRODUCT_IDS
                    );

                    if (empty($productIds)) {
                        $productIds = [];
                    }

                    $productIds[] = $Product->getId();

                    $DownloadEntry->setCustomDataEntry(
                        DownloadEntry::CUSTOM_DATA_KEY_DOWNLOAD_MEMBERSHIP_PRODUCT_IDS,
                        $productIds
                    );
                    $DownloadEntry->update();
                }

                $downloadEntries[] = $DownloadEntry;
            } catch (\Exception $Exception) {
                if ($Exception->getCode() === 404) {
                    QUI\System\Log::writeDebugException($Exception);
                } else {
                    QUI\System\Log::writeException($Exception);
                }
            }
        }

        return $downloadEntries;
    }

    /**
     * Adds a download entry to the database.
     *
     * @param DownloadEntry $DownloadEntry
     * @return int - Download entry id
     *
     * @throws QUI\Database\Exception
     */
    public static function addDownloadEntry(DownloadEntry $DownloadEntry): int
    {
        $DB = QUI::getDataBase();

        $DB->insert(
            self::getTableDownloadEntries(),
            array_merge(
                $DownloadEntry->toDatabaseArray(),
                [
                    'c_date' => date('Y-m-d H:i:s')
                ]
            )
        );

        return (int)$DB->getPDO()->lastInsertId();
    }

    /**
     * Delete download entry with a specific id.
     *
     * @param int $id
     * @return void
     *
     * @throws QUI\Database\Exception
     */
    public static function deleteDownloadEntry(int $id): void
    {
        QUI::getDataBase()->delete(
            self::getTableDownloadEntries(),
            [
                'id' => $id
            ]
        );
    }

    /**
     * @param int $id
     * @return DownloadEntry
     *
     * @throws Exception
     * @throws QUI\Database\Exception
     * @throws QUI\Exception
     */
    public static function getDownloadEntryById(int $id): DownloadEntry
    {
        $result = QUI::getDataBase()->fetch([
            'from' => self::getTableDownloadEntries(),
            'where' => [
                'id' => $id
            ],
            'limit' => 1
        ]);

        if (empty($result)) {
            throw new Exception('DownladEntry with id ' . $id . ' not found.', 404);
        }

        return self::parseDataToDownloadEntry($result[0]);
    }

    /**
     * Get all download entries of a user
     *
     * @param User $User
     * @return DownloadEntry[]
     *
     * @throws QUI\Database\Exception
     * @throws QUI\Exception
     */
    public static function getDownloadEntriesByUser(User $User): array
    {
        $result = QUI::getDataBase()->fetch([
            'from' => self::getTableDownloadEntries(),
            'where' => [
                'user_id' => $User->getId()
            ]
        ]);

        $downloadEntries = [];

        foreach ($result as $row) {
            $downloadEntries[] = self::parseDataToDownloadEntry($row);
        }

        return $downloadEntries;
    }

    /**
     * Parse database data to DownloadEntry
     *
     * @param array $data
     * @return DownloadEntry
     *
     * @throws QUI\Exception
     */
    protected static function parseDataToDownloadEntry(array $data): DownloadEntry
    {
        $User = QUI::getUsers()->get($data['user_id']);
        $DownloadEntry = new DownloadEntry($User, $data['id']);
        $DownloadEntry->setOrderHash($data['order_hash']);

        // Title
        if (!empty($data['title'])) {
            $titleByLang = json_decode($data['title'], true);

            foreach ($titleByLang as $lang => $title) {
                $DownloadEntry->setTitle($lang, $title);
            }
        }

        // Description
        if (!empty($data['description'])) {
            $descriptionByLang = json_decode($data['description'], true);

            foreach ($descriptionByLang as $lang => $description) {
                $DownloadEntry->setDescription($lang, $description);
            }
        }

        // Download items
        if (!empty($data['download_items'])) {
            $downloadItems = json_decode($data['download_items'], true);

            foreach ($downloadItems['quiqqerFiles'] as $quiqqerMediaUrl) {
                try {
                    $DownloadEntry->addQuiqqerMediaItemUrl($quiqqerMediaUrl);
                } catch (\Exception $Exception) {
                    QUI\System\Log::writeException($Exception);
                }
            }

            foreach ($downloadItems['urls'] as $entry) {
                try {
                    $DownloadEntry->addUrl($entry['url'], $entry['title']);
                } catch (\Exception $Exception) {
                    QUI\System\Log::writeException($Exception);
                }
            }
        }

        // Custom data
        if (!empty($data['custom_data'])) {
            $DownloadEntry->setCustomData(json_decode($data['custom_data'], true));
        }

        return $DownloadEntry;
    }

    /**
     * Get module settings
     *
     * @return array
     */
    public static function getSettings(): array
    {
        if (!empty(self::$settings)) {
            return self::$settings;
        }

        try {
            $Conf = QUI::getPackage('quiqqer/user-downloads')->getConfig();
            self::$settings = $Conf->getSection('general');
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }

        return self::$settings;
    }

    /**
     * @return string
     */
    public static function getTableDownloadEntries(): string
    {
        return QUI::getDBTableName(self::TBL_DOWNLOAD_ENTRIES);
    }

    /**
     * @return bool
     */
    public static function canUseDownloadMembershipProducts(): bool
    {
        $PackageManager = QUI::getPackageManager();
        return $PackageManager->isInstalled('quiqqer/erp-plans') && $PackageManager->isInstalled('quiqqer/memberships');
    }

    /**
     * Updates all download files for users that purchases a download membership product.
     *
     * @param DownloadMembershipProductType $Product
     * @return void
     * @throws Exception
     */
    public static function updateProductFilesForDownloadMembership(DownloadMembershipProductType $Product): void
    {
        $FolderField = $Product->getField(self::PRODUCT_FIELD_ID_DOWNLOAD_FILES);

        if (!($FolderField instanceof QUI\ERP\Products\Field\Types\Folder)) {
            return;
        }

        $MediaFolder = $FolderField->getMediaFolder();

        foreach (self::getDownloadEntriesByProduct($Product) as $DownloadEntry) {
            $membershipProductIds = $DownloadEntry->getCustomDataEntry(
                DownloadEntry::CUSTOM_DATA_KEY_DOWNLOAD_MEMBERSHIP_PRODUCT_IDS
            );

            if (empty($membershipProductIds)) {
                continue;
            }

            if (!in_array($Product->getId(), $membershipProductIds)) {
                continue;
            }

            // Remove completely first, then update according to current file active status
            $DownloadEntry->removeMediaItem($MediaFolder);
            $DownloadEntry->addMediaItem($MediaFolder);
            $DownloadEntry->update();
        }
    }

    /**
     * Get all download entries associated with a product.
     *
     * @param DownloadMembershipProductType $Product
     * @return DownloadEntry[]
     */
    public static function getDownloadEntriesByProduct(DownloadMembershipProductType $Product): array
    {
        $MembershipField = MembershipsHandler::getProductMembershipField();

        if (!$MembershipField) {
            return [];
        }

        if (!$Product->hasField($MembershipField->getId())) {
            return [];
        }

        try {
            $membershipId = $Product->getFieldValue($MembershipField->getId());

            if (empty($membershipId)) {
                return [];
            }

            /** @var QUI\Memberships\Membership $Membership */
            $Membership = MembershipsHandler::getInstance()->getChild($membershipId);
            $membershipUserIds = $Membership->getMembershipUserIds();
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return [];
        }

        $productDownloadEntries = [];

        foreach ($membershipUserIds as $membershipUserId) {
            try {
                /** @var QUI\Memberships\Users\MembershipUser $MembershipUser */
                $MembershipUser = MembershipUsersHandler::getInstance()->getChild($membershipUserId);
                $User = $MembershipUser->getUser();

                $downloadEntries = self::getDownloadEntriesByUser($User);

                foreach ($downloadEntries as $DownloadEntry) {
                    $membershipProductIds = $DownloadEntry->getCustomDataEntry(
                        DownloadEntry::CUSTOM_DATA_KEY_DOWNLOAD_MEMBERSHIP_PRODUCT_IDS
                    );

                    if (empty($membershipProductIds)) {
                        continue;
                    }

                    if (!in_array($Product->getId(), $membershipProductIds)) {
                        continue;
                    }

                    $productDownloadEntries[] = $DownloadEntry;
                }
            } catch (\Exception $Exception) {
                QUI\System\Log::writeException($Exception);
            }
        }

        return $productDownloadEntries;
    }

    /**
     * Get all download membership products that are associated with a media folder.
     *
     * @param QUIMediaFolder $Folder
     * @return DownloadMembershipProductType[]
     */
    public static function getDownloadMembershipProductsByFolder(QUIMediaFolder $Folder): array
    {
        $downloadMembershipProducts = ProductsHandler::getProducts([
            'where' => [
                'type' => DownloadMembershipProductType::class
            ]
        ]);

        $products = [];

        /** @var DownloadMembershipProductType $Product */
        foreach ($downloadMembershipProducts as $Product) {
            try {
                /** @var FieldTypeFolder $FolderField */
                $FolderField = $Product->getField(Handler::PRODUCT_FIELD_ID_DOWNLOAD_FILES);
                $ProductDownloadsMediaFolder = $FolderField->getMediaFolder();

                if ($ProductDownloadsMediaFolder && $ProductDownloadsMediaFolder->getId() === $Folder->getId()) {
                    $products[] = $Product;
                }
            } catch (\Exception $Exception) {
                QUI\System\Log::writeException($Exception);
            }
        }

        return $products;
    }

    /**
     * Check if missing files count should be displayed.
     *
     * @return bool
     */
    public static function showMissingFilesCount(): bool
    {
        $setting = self::getSettings();
        return !empty($setting['showMissingFilesCount']);
    }
}
